let gbdir = "."
let version= ref ""
(* pour eviter les warnings! *)
let unix s =
  let r = Unix.system s in
  if r=r then ()
  else ()

let init_trace () =
  unix ("echo \"\" > "^gbdir^"/log_gb")
let trace s =
  unix ("echo \""^s^"\n\" >> "^gbdir^"/log_gb")

open Xml
open Dtd
open Num
open Unix
open Utile
open Entiers
open Polynomes2

let num_0 = Int 0
and num_1 = Int 1
and num_2 = Int 2
and num_10 = Int 10;;

let numdom r =
  let r' = Ratio.normalize_ratio (ratio_of_num r) in
  num_of_big_int(Ratio.numerator_ratio r'),
  num_of_big_int(Ratio.denominator_ratio r');;

(* ------------------------------------------------------------------------- *)
(*  term??                                                                   *)
(* ------------------------------------------------------------------------- *)

type vname = string;;

type term =
| Zero
| Const of Num.num
| Var of vname
| Opp of term
| Add of (term * term)
| Sub of (term * term)
| Mul of (term * term)
| Pow of (term * int);;

let mc = "MUTCONSTRUCT";;
let nc = "noConstr";;
let nu = "uri";;
let ma = "APPLY";;

let turi="cic:/gbtypes/term.ind";;
let tzero = "1";;  (* Zero: term *)
let tconst = "2";; (* Const: Z -> positive -> term *)

let tvar = "3";;   (* Var: positive -> term *)
let topp = "4";;   (* Opp: term *)
let tadd = "5";;   (* Add: term -> term -> term *)
let tsub = "6";;   (* Sub: term -> term -> term *)
let tmul = "7";;   (* Mul: term -> term -> term *)
let tpow = "8";;  (* Pow: term -> positive -> term *)

let luri (* uri *) ="cic:/gbtypes/lineq.ind";;
let lnil = "1";; (* lnil: lineq *)
let lceq = "2";; (* lceq: term -> lineq -> lineq *)


let zuri="cic:/Coq/ZArith/BinInt/Z.ind";;
let z0 = "1";; (* Z0: Z *)
let zpos = "2";; (* xO: positive ->  positive *)
let zneg = "3";; (* xH: positive *)


let puri="cic:/Coq/NArith/BinPos/positive.ind";;
let pxI = "1";; (* xI: positive ->  positive *)
let pxO = "2";; (* xO: positive ->  positive *)
let pxH = "3";; (* xH: positive *)

let mk_elt uri name l = Element(mc, [nu, uri; "noType", "0"; nc, name], l);;

let mk_app uri name l = 
  Element(ma, [], Element(mc, [nu, uri; "noType", "0"; nc, name], [])::l);;

let rec mk_pos n = 
  if n =/ num_1 then  mk_elt puri pxH []
  else if mod_num n num_2 =/ num_0 then
          mk_app puri pxO [mk_pos (quo_num n num_2)]
        else
          mk_app puri pxI [mk_pos (quo_num n num_2)];;

let mk_z z  = 
  if z =/ num_0 then  mk_elt zuri z0 []
  else if z >/ num_0 then
          mk_app zuri zpos [mk_pos z]
        else
          mk_app zuri zneg [mk_pos ((Int 0) -/ z)];;

let rec mk_term t  =  match t with
| Zero -> mk_elt turi tzero []
| Const r -> let (n,d) = numdom r in
               mk_app turi tconst [mk_z n; mk_pos d]     
| Var v -> mk_app turi tvar [mk_pos (num_of_string v)]     
| Opp t1 -> mk_app turi topp [mk_term t1]
| Add (t1,t2) -> mk_app turi tadd [mk_term t1; mk_term t2]
| Sub (t1,t2) -> mk_app turi tsub [mk_term t1; mk_term t2]
| Mul (t1,t2) -> mk_app turi tmul [mk_term t1; mk_term t2]
| Pow (t1,n) ->  if (n = 0) then 
                   mk_app turi tconst [mk_z num_1; mk_pos num_1]     
                 else
                   mk_app turi tpow [mk_term t1; mk_pos (num_of_int n)]
;;

let rec parse_pos p = match p with
| Element (_,_,[p1;p2]) ->
  let a = attrib p1 nc in
  if a = pxO then num_2 */ (parse_pos p2)
             else num_1 +/ (num_2 */ (parse_pos p2))
| _ -> num_1;;

let parse_z z = match z with
| Element (_,_,[p1;p2]) ->
  let a = attrib p1 nc in
  if a = zpos then parse_pos p2 else (num_0 -/ (parse_pos p2))
| _ -> num_0;;

 
let rec parse_term p = match p with
| Element (_,_,[p1;p2]) ->
    let a = attrib p1 nc in
    if a = tvar then Var (string_of_num (parse_pos p2))
    else if a = topp then Opp (parse_term p2)
    else Zero
| Element (_,_,[p1;p2;p3]) ->
    let a = attrib p1 nc in
    if a = tconst then Const ((parse_z p2) // (parse_pos p3))
    else if a = tadd then Add (parse_term p2, parse_term p3)
    else if a = tsub then Sub (parse_term p2, parse_term p3)
    else if a = tmul then Mul (parse_term p2, parse_term p3)
    else if a = tpow then Pow (parse_term p2, int_of_num (parse_pos p3))
    else Zero
| _ -> Zero;;


let rec parse_lineq l1 l2 l3 p = match p with
| Element (_,_,[]) -> ([],[],[])
| Element (_,_,[p1;p2;p3]) ->
  let (l4,l5,l6) = parse_lineq l1 l2 l3 p3 in
  let r = parse_term p2 in
  let a = attrib p1 nc in
  if a = lceq then (r::l4,l5,l6) else
  (l1,l2,l3)
| _ -> (l1,l2,l3);;

let rec parse_request p = 
  version:= attrib p "req";
  trace ("Version: "^(!version));
  match p with
  | Element (_,_,[p1]) -> parse_lineq [] [] [] p1
  | _ -> ([],[],[]);;

(***********************************************************************)
(* Polynomes a coefficients fractionnaires *)
(* un polynome est represent'e par un polynome `a coefficient entiers 
   et un entier (par lequel on divise tous les coefficients *)

type polyf = coef * poly (* denominateur, polynome *)
;;

(* normalise c et les coef de p, rend c positif *)
let norm_polyf (c,p) =
  let a = content p in
  let b = pgcd (abs_coef a) (abs_coef c) in
  let b = if (lt_coef c coef0) then opp_coef b else b in
    (div_coef c b, div_int p b)
;;

let plusPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     plusP (multP (Pint c2) p1) (multP (Pint c1) p2))
;;
let moinsPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     moinsP (multP (Pint c2) p1) (multP (Pint c1) p2))
;;

let multPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     multP p1 p2)
;;

let oppPf (c,p) =
  norm_polyf (c, oppP p)
;;

let puisPf (c,p) n =
 let rec aux n =
   if n=0
  then (coef1, (cf 1))
  else if n=1
  then (c,p)
  else multPf (c,p) (aux (n-1))
  in aux n
;;

let nvars = ref 0;;


(* terme vers couple (c,p) representant le polynome p/c *)

let rec term_polyf t  =  match t with
| Zero -> (coef1, (cf 0))
| Const r -> let (n,d) = numdom r in
   let d = ent_of_string (string_of_num d) in
   let n = ent_of_string (string_of_num n) in
               (d, Pint n)
| Var v -> let n = int_of_string v in
           nvars:= max (!nvars) n;
           (coef1,(x n))
| Opp t1 ->  oppPf (term_polyf t1)
| Add (t1,t2) -> plusPf (term_polyf t1) (term_polyf t2)
| Sub (t1,t2) -> moinsPf (term_polyf t1) (term_polyf t2)
| Mul (t1,t2) -> multPf (term_polyf t1) (term_polyf t2)
| Pow (t1,n) ->  puisPf (term_polyf t1) n
;;

(* polynoem recusrsif vers terme *)

let poly_term p =
  let rec aux p =
  match p with
   |Pint n -> Const (num_of_string (string_of_ent n))
   |Prec (v,coefs) ->
     let res = ref Zero in
     Array.iteri
      (fun i c ->
	 res:=Add(!res, Mul(aux c,
			    Pow (Var (string_of_int v),
				 i))))
      coefs;
     !res
  in aux p;;

(******************************************************************************

  La tactique Gb
*)

let rec pol_rec_to_sparse p v1 v2 =  (* v1 et v2 indices de dbut et de fin *)
  let d = v2-v1+1 in
    match p with
        Pint a ->
          if a= coef0
          then []
          else [a, Array.create (d+1) 0]
      |Prec(v,coefs)->
         let vp=Dansideal.gen d (v2-v+1) in (*j'aurais dit v2-v-1*)
         let res=ref [] in
           for i=(Array.length coefs)-1 downto 0 do
             res:= (Dansideal.plusP d
                      (Dansideal.multP d vp (!res))
                      (pol_rec_to_sparse coefs.(i) v1 v2));
           done;
           !res
;;

(* on inverse l'ordre des variables *)       
let rec pol_sparse_to_rec n1 n2 p =
  List.fold_left (fun r (c,m)->
                    let mp = ref (Pint coef1) in
                      Array.iteri (fun i e ->
                                     if i>0 then
                                       mp:=(multP (!mp) (monome (n2-i+1) e)))
                        m;
                      plusP r (multP (Pint c) (!mp)))
    (Pint coef0)
    p
;;  

let split_regexp rexp s =
  try (Str.split (Str.regexp rexp) s)
  with _ -> [s]

let replace e x s =
  Str.global_replace (Str.regexp e) x s

let rec rang x l =
  match l with
  |[] -> -1
  |y::l -> if x=y then 0 else 1+ (rang x l)

let read_string file =
    let st =open_in file in
    let ls=ref [] in
      (try
        (while true do
	   let c= (input_char st) in
           ls:=(String.make 1 c)::(!ls);
	 done)
       with End_of_file -> () | _ -> ((* pr "probleme dans le fichier";*) ()));
      close_in st;
      String.concat "" (List.rev (!ls))

let read_string2 file =
    let st =open_in file in
    let ls=ref [] in
      (try
        (while true do
	   let s= (input_line st) in
           ls:=(s^"\n")::(!ls);
	 done)
       with End_of_file -> () | _ -> ((* pr "probleme dans le fichier";*) ()));
      close_in st;
      String.concat "" (List.rev (!ls))


(***********************************************************************
  Lecture d' un fichier resultat de GB de JC Faugere, exemple:

#C ./serveur__DMP__Lexp__GINT.exe -pipe2 -read ../inputfgb
#C *** Gb 4.0.1360 for Windows (jcf@calfor.lip6.fr) ***
#C  Evaluation version (only)            
#vars [t,z,x2,x1,e0,e1,e2]

#coeffring integer
#properties [GB, Lexp(7)]

#list
#C Dimension 3   Degree 2
[t-1*z^2*e1-1*z^2*e2-1*z*x2*e0-1*z*x1*e0-1*e0,z*x2*e1-1*z*x1*e2+x2^2*e0-1*x1^2*e0-1*e1+e2,z*x1*e1+z*x1*e2+x2*x1*e0+x1^2*e0-1*e2,x2*e2-1*x1*e1,e0^2,e0*e1,e0*e2,e1^2,e1*e2,e2^2]
#C Time 0.00 sec for computing a GB for a Lexp ordering(Buchberger)

*)


let lire_coefouvarexp s =
(*  trace ("coefouvar "^s);*)
  let (v,e)=
    (match (split_regexp "\\^" s) with
    |[v] -> (v,1)
    |[v;e] -> (v,int_of_string e)
    |_ -> failwith "lire_coefouvarexp" )
  in
  let (c,v,e) =
    try (Str.search_forward (Str.regexp "\\-*[0-9]+") v 0;
	 (int_of_string v, "", 1))
    with _ ->
      if String.sub v 0 1 = "-" 
      then (-1,String.sub v 1 ((String.length v)-1),e)
      else (1,v,e)
  in
(*  trace ("donne c="^(string_of_int c)^", v="^v^", e="^(string_of_int e));*)
  (c,v,e)

let lire_terme d vars s =
  let t = split_regexp "[\\*]" s in
  let lve= List.map lire_coefouvarexp  t in
  let cm = ref coef1 in
  let m = Array.create (d+1) 0 in
  List.iter
    (fun (c,v,e) ->
      cm:= mult_coef (!cm) (coef_of_int c);
      if v<>""
      then m.((rang v vars)+1)<- e
    )
    lve;
  let m = Dansideal.set_deg d m in
  (!cm),m
  

let lire_pol_fgb d vars s =
  let s = replace "-" "+-" s in
  let pol = List.map (lire_terme d vars)  (split_regexp "[\\+]" s) in
  pol


let lire_fgb s =
  if !version="jcf1"
  then
    (let vars = !Dansideal.name_var in
    let d = List.length vars in
    let s = replace "[\n\t\r]" "" s in
    Str.search_forward (Str.regexp "#C Dimension.*#C Time") s 0;
    let s1=Str.matched_string s in
    Str.search_forward (Str.regexp "\\[.+\\]") s1 0;
    let s2=Str.matched_string s1 in
    let s2 = String.sub s2 1 ((String.length s2)-2) in
    let lpol = split_regexp "," s2 in
    List.map (lire_pol_fgb d vars) lpol
    )
  else (
    let vars = !Dansideal.name_var in
    let d = List.length vars in
    trace "lire_fgb commence";
    let s= replace "[\n\t\r]" "" s in
    trace "return vires";
    Str.search_forward (Str.regexp "GROBNER BASIS:.*") s 0;
    let s1=Str.matched_string s in
    trace "GROBNER BASIS trouve";
    Str.search_forward (Str.regexp "\\[.+\\]") s1 0;
    let s2=Str.matched_string s1 in
    let s2 = String.sub s2 1 ((String.length s2)-2) in
    let s2= replace " " "" s2 in
    trace "base trouvee";
    let lpol = split_regexp "," s2 in
    trace "lpol calculee";
(*    trace ("resultat de fgb:\n"^s2);*)
    let res=  List.map (lire_pol_fgb d vars) lpol in
   trace ("res calcule");
    res
   )



(* liste de polynomes lpol=p1,...,pn, 
   echoue ou rend q1,...,qn,q tq 1=q1*p1+...+qn*pn+q*(1-zp)

   
   en general (B.Mourrain):
   exprimer 1=q1p1+...qnpn+q(1-zp)
   calculer la base de grobner de 

   t*(1-zp)-e0, t*p1-e1,...,t*pn-en,ei*ej,t*ei

   avec t>>z>>x1,...,xm>>e0,e1,...,en
   et recuperer dedans un polynome
   t-q*e0-q1*e1-...-qn*en
   Voila.

*)


let deg_nullstellensatz lpol p =
  Random.self_init ();
  let namefile = "fgb_"^(string_of_int (Random.int 65000))
  in
  let inputfgb=gbdir^"/tmp/input"^namefile in
  let outputfgb=gbdir^"/tmp/output"^namefile in

  (* lpol = p1, ..., pn *)
  let n = List.length lpol in
  let i = ref 0 in
  let m= (!nvars) in
    (*  t=m+2 ei=-i z=m+1*)
    
  (* lpol1 = t*p1-e1, ..., t*pn-en *)
  let lpol1 = List.map (fun pi ->
                          i:=!i+1;
                          plusP (multP (x (m+2)) pi) (oppP (x (-(!i)))))
                lpol in
    (* polsup = t*(1-z*p) - e0 *)
  let polsup=(plusP (multP (x (m+2)) (plusP cf1 (oppP (multP (x (m+1)) p)))) (oppP (x 0))) in
  let lpol1=polsup::lpol1 in 
  let leiej=ref [] in
   (* ei*ej *)
    for j=0 to n-1 do
      for k=j+1 to n do
        leiej:=(multP (x (-j)) (x (-k)))::(!leiej);
      done;
    done;

    (* t*ei *)
    for j=0 to n do
      leiej:=(multP (x (-j)) (x (m+2)))::(!leiej);
    done;

    let lpol1=(!leiej)@lpol1 in
    let lvar=ref [] in
    for i=0 to n do lvar:=(!lvar)@["e"^(string_of_int i)^""]; done;
    for i=1 to m do lvar:=["x"^(string_of_int i)^""]@(!lvar); done;
    lvar:=["t1";"z"]@(!lvar);
    Dansideal.name_var:=!lvar;
    let lpol1=List.map (fun p -> pol_rec_to_sparse p (-n-1) (m+2)) lpol1 in
    
    let lpols=
      String.concat "," (List.map Dansideal.stringP lpol1)
    in
    trace ("POLS: "^lpols);

    let gb=
      if !version="lt"
      then  Dansideal.buch (m+n+3) lpol1
      else
	(let textinputfgb =
	  if !version="jcf1"
	  then
	    "#C ./serveur__DMP__Lexp__GINT -pipe2 -read "^inputfgb^"\n"
	    ^"#vars ["^(String.concat "," !lvar)^"]\n"
	    ^"#properties [System]\n"
	    ^"#list\n"
	    ^"["^(String.concat ",\n" (List.map Dansideal.stringP lpol1))^"]\n"
	  else
	    "("^(String.concat "," !lvar)^"),"
	    ^(replace " " ""
		(String.concat ","
		   (List.map Dansideal.stringP lpol1)))
	in
	unix ("echo \""^textinputfgb^"\n\" > "^inputfgb);
	
	let fgbcom=
	  if !version="jcf1"
	  then
	    gbdir^"/serveur__DMP__Lexp__GINT -pipe2 -read "^inputfgb^" >"^outputfgb
	  else
	    gbdir^"/fgbascii <"^inputfgb^" 2>"^outputfgb in
	
	trace "INPUT GB:";
	unix ("cat "^inputfgb^" >>  "^gbdir^"/log_gb");
	unix fgbcom;
	trace "OUTPUT GB:";
	unix ("cat "^outputfgb^" >>  "^gbdir^"/log_gb");
	
	let contenu=read_string2 outputfgb in
	trace "Contenu lu";
	let gb = lire_fgb contenu  in
	trace "Base de grobner lue";
	
	(*unix ("rm -f "^inputfgb);
	unix ("rm -f "^outputfgb);*)
	gb)
    in
      
    (* on garde les polynomes de gb avec t de degre 1 dans le monome dominant
     normalement, il ne doit y en avoir qu'un *)
    let q = (List.filter
	       (fun p -> match p with  
		 [] -> false
	       |(_,mon)::_ -> mon.(1)=1) 
               gb) in
    let q' = List.nth q 0 in 
    let lq = ref [] in
    let degz = ref 0 in           
    for i=1 to n+1 do
      let i0=n+m+4-i in
      (* les monomes qui sont construits a l'aide de ei *)
      let qi=List.filter
	  (fun p ->
	    match p with
              (_,mon) ->  mon.(i0)>0 
	  ) 
          q' in 
      (* mutiliplier par le coef de e (1 ou -1) dans q' *)
      let qi = List.map
	  (fun p ->
	    match p with 
              (a,mon) -> 
		((mult_coef a 
		    (opp_coef
		       (coef_of_int
			  (signe_coef (fst (List.nth q' 0)))))),mon))
	  qi in
      (* on enleve ei *)
      let qqi = List.map
	  (fun p ->
            let mon =snd p in
            let c = fst p in
            let mon1=Array.copy mon in
            mon1.(i0)<-mon.(i0)-1;
            (c,mon1))
	  qi in
      (* et on regarde le degre de z *)
      List.iter
	(fun qi ->
          let deg' = (snd qi).(2) in
          (if (compare deg' !degz)=1 then 
	    degz:=deg'))
	qqi;
      lq:=[qqi]@(!lq);
    done;

    let pi=pol_rec_to_sparse p (-n-1) (m+2) in
    lq := List.map
	(fun hi ->
	  (* z^a devient z^(degz-a) *)
	  let gi = Dansideal.specialsubstP (m+3) hi 2 !degz in
	  (*z devient p*)
	  Dansideal.substP (m+3) gi 2 pi)
	!lq;

    let lqr=List.map (fun p->pol_sparse_to_rec (-n-1) (m+2) p) !lq in
    
    let specialcoef = fst (List.nth q' 0) in
    let coefs=
      String.concat ",\n" (List.map string_of_P lqr)
    in
    trace ("degre: "^(string_of_int !degz)^"\n"
		 ^"coefficients:\n "^coefs);
    trace ("ce polynome doit etre nul:"
	     ^"("^(string_of_P p)^")^"^(string_of_int !degz)
	   ^"\n-("
	   ^(String.concat "\n+"
	       (List.map2
		  (fun c p -> "("^(string_of_P c)^")*("^(string_of_P p)^")")
		  (List.tl lqr) lpol))
	       ^")");
    (specialcoef,(lqr,!degz))
;;                

(* calcul de l'exposant et des coefficients de la combinaison lineaire *)

let comb_lin1 lp =
  nvars:=0;(* mise a jour par term_polyf *)
  let lp0 = List.map term_polyf lp in 
  let lp = List.map snd lp0 in
  (* lp = liste de polynomes recursifs *)
  let p::lp1 = List.rev lp in
  let lpol = List.rev lp1 in
(* montrer lp = 0 => p=0 *)
  let (c,(_::lq,d))=  deg_nullstellensatz lpol p in
  let c = abs_ent c in
(* on a
         (-c)*p^d = sum lq*lp
*)
  let lc = (Pint c)::lq in
  (Pow ((poly_term p),d))::(List.map poly_term lc)





